# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import argparse
import pathlib

import h5py
import torch
import torchvision.transforms.functional as tf


def save_images(tensors_dict, output_dir):
    for key, value in tensors_dict.items():
        output_file = output_dir / pathlib.Path(key + ".png")
        tf.to_pil_image(value.float()).save(output_file)
        print(f"Saved file {output_file}")


def extract_images_from_hdf5(hdf5_file):
    tensors_dict = {}
    with h5py.File(hdf5_file, "r") as f:
        img = f["image"]
        assert isinstance(img, h5py.Dataset)
        mask = f["mask"]
        assert isinstance(mask, h5py.Dataset)
        tensors_dict["mask"] = torch.from_numpy(mask[:])
        tensors_dict["image"] = torch.from_numpy(img[:])
        prediction = f["prediction"]
        assert isinstance(prediction, h5py.Dataset)
        tensors_dict["prediction"] = torch.from_numpy(prediction[:])
        target = f["target"]
        assert isinstance(target, h5py.Dataset)
        tensors_dict["target"] = torch.from_numpy(target[:])

    return tensors_dict


def get_hdf5_files(directory) -> list[pathlib.Path]:
    return directory.glob("**/*.hdf5")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_dir",
        type=pathlib.Path,
        help="Directory in which hdf5 files are located.",
    )

    args = parser.parse_args()

    hdf5_files = get_hdf5_files(args.input_dir)
    for hdf5_file in hdf5_files:
        tensors_dict = extract_images_from_hdf5(hdf5_file)

        save_dir = pathlib.Path(hdf5_file).with_suffix("").with_suffix("")
        save_dir.mkdir(parents=True, exist_ok=True)
        save_images(tensors_dict, save_dir)


if __name__ == "__main__":
    main()
