# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Script to automatically generate histograms from DataLoader samples.

!!!DO NOT USE THIS TO "FIX" TESTS!!!
If a test fails because the expected histogram is different than the reference
histogram generated by this, make sure it is not an issue with the code first.
A valid reason to re-generate histograms would be a change in the applied transforms.
"""

import importlib
import json
import pathlib
import random
import re
from typing import TypeAlias

import click
import matplotlib.pyplot as plt
import numpy
import PIL
import torch
from clapper.click import ConfigCommand as _BaseConfigCommand
from clapper.click import ResourceOption
from clapper.logging import setup
from mednet.data.typing import Dataset
from mednet.scripts.database import _get_raw_databases
from torchvision.transforms.functional import to_pil_image

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")

Histogram: TypeAlias = list[int]


class ConfigCommand(_BaseConfigCommand):
    """A click command-class that has the properties of
    :py:class:`clapper.click.ConfigCommand` and adds verbatim epilog
    formatting.
    """

    def format_epilog(
        self, _: click.core.Context, formatter: click.formatting.HelpFormatter
    ) -> None:
        """Format the command epilog during --help.

        Parameters
        ----------
            _
                The current parsing context.
            formatter
                The formatter to use for printing text.
        """

        if self.epilog:
            formatter.write_paragraph()
            for line in self.epilog.split("\n"):
                formatter.write(line + "\n")


def prettify_json(data: str):
    """Prettify function, in order to reduce file size and not having to rely on external tools.

    Parameters
    ----------
    data
        The json data to prettify.

    Returns
    -------
        The prettified json data.
    """
    prettified = json.dumps(data)
    # Opening bracket
    prettified = re.sub(r'{"', '{\n  "', prettified)
    # Carriage return after dictionary key
    prettified = re.sub(r'": \[', '": [\n    ', prettified)
    # Carriage returns before new dictionary key
    prettified = re.sub(r'\]\], "', ']\n  ],\n  "', prettified)
    # Carriage return after end of sample
    prettified = re.sub(r'\], \["', '],\n    ["', prettified)
    # Ending brackets
    return re.sub(r"\]}", "\n  ]\n}\n", prettified)


def display_histogram(sample_path: pathlib.Path, image: PIL.Image, histogram):
    color_values = range(256)

    _, ax = plt.subplot_mosaic([["tbx", "histogram"]], figsize=(7, 3.5))

    ax["tbx"].imshow(image, cmap="gray", vmin=0, vmax=255)
    ax["tbx"].axis("off")

    ax["histogram"].plot(color_values, histogram)

    print(sample_path)
    plt.show()
    # import pdb; pdb.set_trace()


def compute_histogram_from_image_tensor(
    image_tensor: torch.Tensor,
) -> Histogram:
    """Compute the histogram from a Tensor containing image data.

    Parameters
    ----------
    image_tensor
        The tensor from which to compute the histogram. Expected size [num_channels, width, height].

    Returns
    -------
     The computed histogram, of length num_channels * width * height. The channels are in RGB order.
    """
    channels_histograms = []
    for color_channel in image_tensor:
        color_channel = numpy.multiply(color_channel.numpy(), 255).astype(int)
        histogram = numpy.histogram(color_channel, bins=256, range=(0, 256))
        # Note that the image channels are concatenated
        channels_histograms.extend(histogram[0].tolist())

    return channels_histograms


def get_histograms_for_split(
    data: Dataset,
    num_histograms: int,
    images_subdir: pathlib.Path,
    debug: bool = False,
) -> list[tuple[pathlib.Path, Histogram]]:
    """Compute the histograms for samples in a database split and saves the samples after transformations have been applied for manual verification.

    Parameters
    ----------
        data
            A mednet Dataset instance.
        num_histograms
            The number of histograms to generate per split.
            Setting a value of 0 will conpute a histogram for every sample in the split.
        images_subdir
            Directory in which to save the samples with transforms applied.
        debug
            If set to True, will open a window with the transformed image and corresponding histogram for each sample.

    Returns
    -------
        A list of path to samples and corresponding histograms.
    """
    computed_histograms = []
    sample_indexes: list[int] = []

    # Compute histograms for entire dataset if num_histograms is set to 0
    if num_histograms == 0:
        sample_indexes = list(range(len(data)))
    else:
        for _ in range(num_histograms):
            sample_indexes.append(random.randint(0, len(data) - 1))

    for index in sample_indexes:
        current_sample = data[index]

        image_data = current_sample[0]

        if isinstance(image_data, dict):
            image_data = image_data["image"]

        sample_path = current_sample[1]["name"]
        histogram = compute_histogram_from_image_tensor(image_data)

        image_path = images_subdir / sample_path
        if not image_path.parent.exists():
            image_path.parent.mkdir(parents=True, exist_ok=True)

        img = to_pil_image(image_data)
        img.save(image_path.with_suffix(".png"))

        computed_histograms.append((sample_path, histogram))

        if debug:
            display_histogram(sample_path, img, histogram)

    return computed_histograms


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
)
@click.option(
    "--model",
    "-m",
    help="""A lightning module instance implementing the network architecture
    (not the weights, necessarily) to be used for prediction.""",
    required=False,
    cls=ResourceOption,
)
@click.option(
    "--database",
    "-d",
    help="""The name of the database to use for generating the histograms. If not specified, tries all available databases.""",
    required=False,
    type=str,
    cls=ResourceOption,
)
@click.option(
    "--library",
    "-l",
    help="""The library in which to look for the model and database.""",
    required=True,
    type=str,
    cls=ResourceOption,
)
@click.option(
    "--fold-name",
    "-f",
    help="""The name of the fold to use for generating the histograms.""",
    required=False,
    default="default",
    type=str,
    cls=ResourceOption,
)
@click.option(
    "--num-histograms",
    "-n",
    help="""The number of histograms to generate per split.""",
    required=False,
    default=5,
    type=int,
    cls=ResourceOption,
)
@click.option(
    "--seed",
    "-s",
    help="""The seed number to initialize the RNG with.""",
    required=False,
    default=42,
    type=int,
    cls=ResourceOption,
)
@click.option(
    "--debug",
    help="""If true, opens a window with the sample and the histogram(s) side by side after each computation.""",
    required=False,
    type=bool,
    cls=ResourceOption,
)
def main(
    model,
    database,
    library,
    num_histograms,
    fold_name,
    seed,
    debug,
    **_,
):
    random.seed(seed)

    databases_to_use = {}
    if database is not None:
        databases_to_use[database] = _get_raw_databases()[database]
    else:
        databases_to_use = _get_raw_databases()

    model_transforms = []

    model_name = ""
    if model is not None:
        model_name = f"_{model.name}"
        model_transforms = model.model_transforms
        print(f"Loaded model transforms from {model.name}")

    for k, v in databases_to_use.items():
        print(f"Processing database {k}")

        if "datadir" not in v:
            # this database does not have a "datadir"
            continue

        if v["datadir"] is not None:
            try:
                datamodule = importlib.import_module(
                    f".{fold_name}", f"mednet.{library}.config.data.{k}"
                ).datamodule
            except ModuleNotFoundError as e:
                print(e)
                continue

            output_dir = pathlib.Path(
                f"histograms_{library}{model_name}_{k}_{fold_name}"
            )
            images_subdir = output_dir / "images"
            if not output_dir.exists():
                output_dir.mkdir(parents=True, exist_ok=True)

            output_file = output_dir / f"histograms{model_name}_{k}_{fold_name}.json"

            output_json_data = {}

            datamodule.model_transforms = model_transforms
            datamodule.setup("predict")

            for split_name, dataset_data in datamodule._datasets.items():  # noqa: SLF001
                output_json_data[split_name] = get_histograms_for_split(
                    dataset_data, num_histograms, images_subdir, debug
                )

            output_json_data = prettify_json(output_json_data)

            with output_file.open("w") as o_f:
                o_f.write(output_json_data)
                print(f"Saved histogram for {k} at {output_file}")
        else:
            print(f"Skipping {k} as database is not installed.")


if __name__ == "__main__":
    main()
