# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Resizes the Montgomery dataset to speed-up processing.

The chosen method follows the same strategy for loading and preprocessing this
dataset.
"""

import pathlib
import sys

import numpy
import PIL.Image
from mednet.data.image_utils import remove_black_borders
from mednet.models.transforms import square_center_pad
from torch.nn.functional import interpolate
from torchvision.transforms.functional import resize, to_pil_image, to_tensor


def main():
    if len(sys.argv) != 2:
        print(__doc__)
        print(f"Usage: python3 {sys.argv[0]} <root-of-montgomery-dataset>")
        sys.exit(0)

    montgomery_root = pathlib.Path(sys.argv[1])

    base_image_dir = pathlib.Path("CXR_png")
    left_mask_dir = pathlib.Path("ManualMask") / "leftMask"
    right_mask_dir = pathlib.Path("ManualMask") / "rightMask"

    for f in (montgomery_root / base_image_dir).glob("*.png"):
        print(f.name)
        image = PIL.Image.open(f)
        image, mask = remove_black_borders(image)
        image = to_tensor(image)
        image = square_center_pad(image)
        image = resize(image, 512)
        image = to_pil_image((255 * image).byte(), "L")
        base_image_dir.mkdir(parents=True, exist_ok=True)
        image.save(base_image_dir / f.name)

        left_mask = PIL.Image.open(montgomery_root / left_mask_dir / f.name)
        left_mask = PIL.Image.fromarray(
            numpy.asarray(left_mask)[numpy.ix_(mask.any(1), mask.any(0))],
        )
        left_mask = to_tensor(left_mask)
        left_mask = square_center_pad(left_mask)
        left_mask = interpolate(left_mask.unsqueeze(0), 512, mode="nearest").squeeze(0)
        left_mask = to_pil_image((255 * left_mask).byte(), "L")
        left_mask = left_mask.convert("1")
        left_mask_dir.mkdir(parents=True, exist_ok=True)
        left_mask.save(left_mask_dir / f.name)

        right_mask = PIL.Image.open(montgomery_root / right_mask_dir / f.name)
        right_mask = PIL.Image.fromarray(
            numpy.asarray(right_mask)[numpy.ix_(mask.any(1), mask.any(0))],
        )
        right_mask = to_tensor(right_mask)
        right_mask = square_center_pad(right_mask)
        right_mask = interpolate(right_mask.unsqueeze(0), 512, mode="nearest").squeeze(
            0
        )
        right_mask = to_pil_image((255 * right_mask).byte(), "L")
        right_mask = right_mask.convert("1")
        right_mask_dir.mkdir(parents=True, exist_ok=True)
        right_mask.save(right_mask_dir / f.name)


if __name__ == "__main__":
    main()
